# -*- coding: utf-8 -*-
from __future__ import absolute_import
import json
from dataclasses import asdict

from redis import Redis

from dqueue.exceptions import ItemExistError, ItemKeyError
from dqueue.item import Item
from dqueue.storages.base import BaseStorage
from dqueue.utils import after_timestamp, datetime_to_timestamp


class RedisStorage(BaseStorage):

    def __init__(self, db=0, queue_key="delay:queue", due_times_key="delay:due_times", **connect_args):
        self.queue_key = queue_key
        self.due_times_key = due_times_key
        self.redis = Redis(db=int(db), **connect_args)

    def get_due_items(self, now):
        now_timestamp = datetime_to_timestamp(now)
        ids = self.redis.zrangebyscore(self.due_times_key, 0, now_timestamp)
        if ids:
            states = self.redis.hmget(self.queue_key, *ids)
            return [Item(**json.loads(state)) for state in states]
        return []

    def get_all_items(self):
        states = self.redis.hgetall(self.queue_key)
        return [Item(**json.loads(v)) for k, v in states.items()]

    def add_item(self, item: Item, ttl):
        if self.redis.hexists(self.queue_key, item.id):
            raise ItemExistError(item.id)

        with self.redis.pipeline() as pipe:
            pipe.multi()
            pipe.hset(self.queue_key, item.id, json.dumps(asdict(item)))
            pipe.zadd(self.due_times_key, {item.id: after_timestamp(ttl)})
            pipe.execute()

    def remove_item(self, item):
        item_id = item if not isinstance(item, Item) else item.id
        if not self.redis.hexists(self.queue_key, item_id):
            raise ItemKeyError(item_id)

        with self.redis.pipeline() as pipe:
            pipe.hdel(self.queue_key, item_id)
            pipe.zrem(self.due_times_key, item_id)
            pipe.execute()

    def remove_all_items(self):
        with self.redis.pipeline() as pipe:
            pipe.delete(self.queue_key)
            pipe.delete(self.due_times_key)
            pipe.execute()

    def shutdown(self):
        self.redis.connection_pool.disconnect()



# -*- coding: utf-8 -*-
from queue import Queue, Empty
from typing import Union

from dqueue.queues.base import BaseQueue


class MemoryQueue(BaseQueue):
    def __init__(self):
        super().__init__()
        self.q = Queue()
        self.start()

    def _get(self, block: Union[bool, int] = True):
        _block = True
        _timeout = None
        if isinstance(block, bool):
            _block = block
        elif isinstance(block, int):
            _block = True
            _timeout = block
        else:
            raise ValueError("block must be bool or int")

        try:
            return self.q.get(block=_block, timeout=_timeout)
        except Empty:
            return None

    def get(self, count=1, block: Union[bool, int] = True):
        for _ in range(count):
            item = self._get(block=block)
            if item is None:
                break
            yield item

    def push(self, item):
        self.q.put(item)

    def shutdown(self):
        super(MemoryQueue, self).shutdown()



# -*- coding: utf-8 -*-
"""
基于Redis的延迟队列

    基于Redis的延迟队列，可以将任务放入队列，并且可以设置延迟时间，当达到延迟时间时，任务将被取出。
    本模式是采用Redis的Stream实现，因此在使用本模式前，必须达到的基本要求是Redis的版本应该是5.x及以上。

"""
from __future__ import absolute_import
from dataclasses import asdict
from typing import Union

from redis import Redis

from dqueue.queues.base import BaseQueue


class RedisQueue(BaseQueue):
    def __init__(self, db=0, group_key="default_group", queue_key="delay:stream", streamId="0", **connect_args):
        super().__init__()
        self.group_key = group_key
        self.queue_key = queue_key
        self.streamId = streamId
        self.redis = Redis(db=int(db), **connect_args)

        self._create_xgroup()
        self.start()

    def _create_xgroup(self):
        try:
            self.redis.xgroup_create(self.queue_key, self.group_key, id=self.streamId, mkstream=True)
        except:
            pass

    def push(self, item):
        self.redis.xadd(self.queue_key, asdict(item))

    def get(self, count=1, consumer="default", streamId=">", block=None):
        """
        Read from a stream via a consumer group;
        :param count: if set, only return this many items, beginning with the
               earliest available;
        :param consumer: name of the requesting consumer;
        :param streamId: IDs indicate the last ID already seen;
        :param block: number of milliseconds to wait, if nothing already present.
        :return:
        """
        res = self.redis.xreadgroup(self.group_key, consumer, {self.queue_key: streamId}, count=count, block=block)
        if not res:
            return []
        [[_, messages]] = res
        return messages

    def ack(self, message: Union[list, tuple, bytes, str]):
        if not len(message):
            return
        if isinstance(message, (bytes, str)):
            ids = [message]
        elif isinstance(message, list):
            ids = [msg_id for msg_id, _ in message]
        elif isinstance(message, tuple):
            ids = [message[0]]
        else:
            raise TypeError("message must be bytes, str, list or tuple")
        self.redis.xack(self.queue_key, self.group_key, *ids)

    def shutdown(self):
        super(RedisQueue, self).shutdown()
        self.redis.connection_pool.disconnect()


# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from datetime import datetime
from threading import RLock

from dqueue.runners import BackgroundRunner, TIMEOUT_MIN
from dqueue.storages.base import BaseStorage
from dqueue.storages.redis import RedisStorage
from dqueue.storages.memory import MemoryStorage


class BaseQueue(ABC):
    # TODO: 未来自动引入，暂时手动
    _storage_classes = {
        'redis': RedisStorage,
        'memory': MemoryStorage
    }

    def __init__(self):
        self._storages = {}
        self._storages_lock = self._create_lock()
        self._runner = None

    @abstractmethod
    def get(self):
        ...

    @abstractmethod
    def push(self, item):
        ...

    def add(self, item, ttl, storage='default'):
        storage: BaseStorage = self._lookup_storage(storage)
        storage.add_item(item, ttl)

    def _lookup_storage(self, alias):
        try:
            return self._storages[alias]
        except KeyError:
            raise KeyError('No lookup storage: %s' % alias)

    def _process(self):
        now = datetime.now()
        with self._storages_lock:
            for storage_alias, storage in self._storages.items():
                due_items = storage.get_due_items(now)
                for item in due_items:
                    self.push(item)
                    storage.remove_item(item)
        return TIMEOUT_MIN

    def add_storage(self, storage, alias='default', **options):

        if isinstance(storage, BaseStorage):
            self._storages[alias] = storage
        elif isinstance(storage, str):
            self._storages[alias] = self._create_plugin_instance(storage, options)
        else:
            raise TypeError('storage must be BaseStorage or str')

        self._runner.wakeup()

    def _create_plugin_instance(self, alias, options):
        class_container, base_class = (self._storage_classes, BaseStorage)
        plugin_cls = class_container[alias]
        return plugin_cls(**options)

    def start(self, paused=False):
        self._runner = self._create_default_runner()
        self._runner.start(self)

        with self._storages_lock:
            if 'default' not in self._storages:
                self.add_storage(self._create_default_storage(), 'default')

            for alias, storage in self._storages.items():
                storage.start(self, alias)

        if not paused:
            self._runner.wakeup()

    def _create_lock(self): # noqa
        return RLock()

    def _create_default_storage(self):  # noqa
        return MemoryStorage()

    def _create_default_runner(self):  # noqa
        return BackgroundRunner()

    @abstractmethod
    def shutdown(self):
        for storage in self._storages.values():
            storage.shutdown()
        self._runner.shutdown()

    def __repr__(self):
        return '<%s>' % self.__class__.__name__



# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod


class BaseStorage(ABC):

    @abstractmethod
    def get_due_items(self, now):
        """
        返回到期的数据
        :return: list[item]
        """

    @abstractmethod
    def get_all_items(self):
        """
        返回数据库中的所有数据
        :return: list[item]
        """

    @abstractmethod
    def add_item(self, item, ttl):
        """
        添加数据
        :param item: Item Data
        :param ttl: 毫秒 过期时间
        :return:
        """

    @abstractmethod
    def remove_item(self, item):
        """
        删除数据
        :param item: Item Data
        :return:
        """

    @abstractmethod
    def remove_all_items(self):
        """
        删除数据库中所有的数据
        :return:
        """

    def shutdown(self):
        """用于释放资源"""

    def start(self, *args, **kwargs): ...

    def __repr__(self):
        return '<%s>' % self.__class__.__name__



# -*- coding: utf-8 -*-
from dqueue.exceptions import ItemExistError, ItemKeyError
from dqueue.item import Item
from dqueue.storages.base import BaseStorage
from dqueue.utils import after_timestamp, datetime_to_timestamp


class MemoryStorage(BaseStorage):

    def __init__(self):
        self._items = []
        self._item_index = {}

    def get_due_items(self, now):
        now_timestamp = datetime_to_timestamp(now)
        return [
            item
            for item, timestamp in self._items
            if timestamp and timestamp <= now_timestamp
        ]

    def get_all_items(self):
        return [item[0] for item in self._items]

    def add_item(self, item: Item, ttl):
        if item.id in self._item_index:
            raise ItemExistError(item.id)
        timestamp = after_timestamp(ttl)
        index = self._get_item_index(timestamp, item.id)
        self._items.insert(index, (item, timestamp))
        self._item_index[item.id] = (item, timestamp)

    def remove_item(self, item):
        item_id = item if not isinstance(item, Item) else item.id

        item, timestamp = self._item_index.get(item_id, (None, None))
        if item is None:
            raise ItemKeyError(item_id)
        index = self._get_item_index(timestamp, item_id)
        del self._items[index]
        del self._item_index[item_id]

    def remove_all_items(self):
        self._items = []
        self._item_index = {}

    def shutdown(self):
        self.remove_all_items()

    def _get_item_index(self, timestamp, job_id):
        """二分查找法"""
        lo, hi = 0, len(self._items)
        timestamp = float('inf') if timestamp is None else timestamp
        while lo < hi:
            mid = (lo + hi) // 2
            mid_job, mid_timestamp = self._items[mid]
            mid_timestamp = float('inf') if mid_timestamp is None else mid_timestamp
            if mid_timestamp > timestamp:
                hi = mid
            elif mid_timestamp < timestamp:
                lo = mid + 1
            elif mid_job.id > job_id:
                hi = mid
            elif mid_job.id < job_id:
                lo = mid + 1
            else:
                return mid

        return lo



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['dqueue', 'dqueue.queues', 'dqueue.storages']

package_data = \
{'': ['*']}

install_requires = \
['AMQPStorm>=2.10.5,<3.0.0', 'redis>=4.3.4,<5.0.0']

setup_kwargs = {
    'name': 'dqueue',
    'version': '0.1.2',
    'description': 'python通用延迟消息队列',
    'long_description': "\n## Usage\n\n> pip install dqueue\n\n## Example\n### Memory Queue\n\n```python\n# memory queue\nimport threading\n\nfrom dqueue.queues import MemoryQueue\nfrom dqueue.item import Item\n\nqueue = MemoryQueue()\n\n# 模拟不同延迟发布消息\ndef publisher():\n    for i in range(10):\n        item = Item(data=f'hello {i}')\n        queue.add(item, i * 1000)\n\n# 模拟消费端取延迟消息\ndef consumer():\n    while True:\n        messages = queue.get(block=10 * 1000)\n        for msg in messages:\n            print('message:', msg)\n\n\nthreading.Thread(target=publisher).start()\nthreading.Thread(target=consumer).start()\n```\n\n### Redis Queue\n\nYou can continue with the example above and Just change the declare.\n\n```python\nfrom dqueue.queues import RedisQueue\n\nqueue = RedisQueue()\n\n# same code\n```",
    'author': 'miclon',
    'author_email': 'jcnd@163.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mic1on/delay_queue',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
import time
from datetime import datetime, timedelta


def after_timestamp(milliseconds=0):
    after_datetime = datetime.now() + timedelta(milliseconds=milliseconds)
    return datetime_to_timestamp(after_datetime)


def datetime_to_timestamp(datetimeval: datetime):
    timetuple = datetimeval.timetuple()
    return int(round(time.mktime(timetuple) * 1000) + datetimeval.microsecond/1000)



# -*- coding: utf-8 -*-
class ItemExistError(Exception):
    """数据已存在"""

    def __init__(self, item_id):
        super(ItemExistError, self).__init__(u'item %s was exist' % item_id)


class ItemKeyError(KeyError):
    def __init__(self, item_id):
        super(ItemKeyError, self).__init__(u'item %s not found' % item_id)




# -*- coding: utf-8 -*-
from threading import Thread, Event
import _thread

TIMEOUT_MAX = _thread.TIMEOUT_MAX
TIMEOUT_MIN = 1


class BlockingRunner:
    _event = None
    _queue = None

    def start(self, queue, *args, **kwargs):
        self._queue = queue
        if self._event is None or self._event.is_set():
            self._event = Event()
        self._main_loop()

    def shutdown(self, *args, **kwargs):
        self._event.set()

    def _main_loop(self):
        wait_seconds = TIMEOUT_MAX
        while True:
            self._event.wait(wait_seconds)
            self._event.clear()
            wait_seconds = self._queue._process()  # noqa

    def wakeup(self):
        self._event.set()


class BackgroundRunner(BlockingRunner):
    def start(self, queue, *args, **kwargs):
        self._queue = queue
        if self._event is None or self._event.is_set():
            self._event = Event()

        self._thread = Thread(target=self._main_loop, name='delay_queue')
        self._thread.start()

    def shutdown(self, *args, **kwargs):
        super(BackgroundRunner, self).shutdown(*args, **kwargs)
        self._thread.join()
        del self._thread



# -*- coding: utf-8 -*-
"""
RabbitMQ 延迟队列需要自行安装插件：
>> https://github.com/rabbitmq/rabbitmq-delayed-message-exchange

RabbitMQ 延迟队列使用教程：
@see https://52caiji.com/posts/other/rabbitmq-delay-queue.html
"""
import json
from dataclasses import asdict
from typing import Union, List

from amqpstorm import Connection, Message

from dqueue.queues.base import BaseQueue


class RabbitMQQueue(BaseQueue):
    def __init__(self,
                 exchange_name="delayed.exchange",
                 queue_name="delayed.queue",
                 routing_key="delayed.routing.key",
                 hostname="localhost",
                 username="guest",
                 password="guest",
                 **connect_args):
        super().__init__()
        self.exchange_name = exchange_name
        self.queue_name = queue_name
        self.routing_key = routing_key
        self.conn = Connection(hostname, username, password, **connect_args)
        self.channel = self.conn.channel()

        self._configuration()

    def _configuration(self):
        arguments = {"x-delayed-type": "direct"}
        self.channel.exchange.declare(self.exchange_name,
                                      exchange_type="x-delayed-message",  # plugin support
                                      durable=True,
                                      arguments=arguments)
        self.channel.queue.declare(self.queue_name, durable=True)
        self.channel.queue.bind(
            queue=self.queue_name,
            exchange=self.exchange_name,
            routing_key=self.routing_key,
        )

    def add(self, item, ttl, storage='default'):
        self.channel.basic.publish(
            json.dumps(asdict(item)),
            self.routing_key,
            self.exchange_name,
            properties={'headers': {"x-delay": ttl}}
        )

    def get(self, count=1, *args, **kwargs) -> List[Union[Message, None]]:
        messages = []
        for _ in range(count):
            message = self.channel.basic.get(self.queue_name, *args, **kwargs)
            if not message:
                continue
            messages.append(message)
        return messages

    def push(self, item): ...

    def ack(self, message):
        if isinstance(message, Message):
            message.ack()
        elif isinstance(message, int):
            self.channel.basic.ack(delivery_tag=message)
        else:
            raise ValueError("message must be Message or int")

    def shutdown(self):
        super(RabbitMQQueue, self).shutdown()
        self.channel.close()
        self.conn.close()
